/*===============================================================================================
 codec_sid.dll
 Ver. 1.2
 Copyright (c), Blazer Studios, 2007.
 Visit http://andreas.blazer.nu for latest version
 Built with Sidplay2, http://sidplay2.sourceforge.net
 
===============================================================================================*/
#define HAVE_MSWINDOWS

#include "fmod_errors.h"
#include <iostream>
#include "sidplay2.h"
#include "resid.h"


void ERRCHECK(FMOD_RESULT result)
{
    if (result != FMOD_OK)
    {
        printf("FMOD error! (%d) %s\n", result, FMOD_ErrorString(result));
        exit(-1);
    }
}

FMOD_RESULT F_CALLBACK sidopen(FMOD_CODEC_STATE *codec, FMOD_MODE usermode, FMOD_CREATESOUNDEXINFO *userexinfo);
FMOD_RESULT F_CALLBACK sidclose(FMOD_CODEC_STATE *codec);
FMOD_RESULT F_CALLBACK sidread(FMOD_CODEC_STATE *codec, void *buffer, unsigned int size, unsigned int *read);
FMOD_RESULT F_CALLBACK sidsetposition(FMOD_CODEC_STATE *codec, int subsound, unsigned int position, FMOD_TIMEUNIT postype);
FMOD_RESULT F_CALLBACK sidgetlength(FMOD_CODEC_STATE *codec, unsigned int *length, FMOD_TIMEUNIT lengthtype);
FMOD_RESULT F_CALLBACK sidgetposition(FMOD_CODEC_STATE *codec, unsigned int *position, FMOD_TIMEUNIT postype);


FMOD_CODEC_DESCRIPTION sidcodec =
{
    "FMOD Sid player plugin",			// Name.
    0x00012000,                         // Version 0xAAAABBBB   A = major, B = minor.
    1,                                  // Force everything using this codec to be a stream
    FMOD_TIMEUNIT_MS|FMOD_TIMEUNIT_MODPATTERN,					// The time format we would like to accept into setposition/getposition.
    &sidopen,                           // Open callback.
    &sidclose,                          // Close callback.
    &sidread,                           // Read callback.
    &sidgetlength,                      // Getlength callback.  (If not specified FMOD return the length in FMOD_TIMEUNIT_PCM, FMOD_TIMEUNIT_MS or FMOD_TIMEUNIT_PCMBYTES units based on the lengthpcm member of the FMOD_CODEC structure).
    &sidsetposition,                    // Setposition callback.
    &sidgetposition,                                  // Getposition callback. (only used for timeunit types that are not FMOD_TIMEUNIT_PCM, FMOD_TIMEUNIT_MS and FMOD_TIMEUNIT_PCMBYTES).
    0                                   // Sound create callback (don't need it)
};
class sid : public sidplay2
{
    FMOD_CODEC_STATE *_codec;
    int _pos;

	public:
		sid(FMOD_CODEC_STATE *codec)
		{
			_codec = codec;
			_pos = 0;
			tune = NULL;
			
			memset(&sidwaveformat, 0, sizeof(sidwaveformat));
		}

		~sid()
		{
			delete tune;
		}
		SidTune *tune;
		int subsongs;
		ReSIDBuilder *rs;
		sid2_config_t m_cfg;
		FMOD_CODEC_WAVEFORMAT sidwaveformat;
	
};

/*
    FMODGetCodecDescription is mandatory for every fmod plugin.  This is the symbol the registerplugin function searches for.
    Must be declared with F_API to make it export as stdcall.
    MUST BE EXTERN'ED AS C!  C++ functions will be mangled incorrectly and not load in fmod.
*/
#ifdef __cplusplus
extern "C" {
#endif

F_DECLSPEC F_DLLEXPORT FMOD_CODEC_DESCRIPTION * F_API FMODGetCodecDescription()
{
    return &sidcodec;
}

#ifdef __cplusplus
}
#endif



FMOD_RESULT F_CALLBACK sidopen(FMOD_CODEC_STATE *codec, FMOD_MODE usermode, FMOD_CREATESOUNDEXINFO *userexinfo)
{   
	FMOD_RESULT       result;

	sid *m = new sid(codec);
	
	char* smallBuffer;
	smallBuffer = new char[4];

	

	unsigned int *bytesread;
	bytesread = new unsigned int;

	result = codec->fileseek(codec->filehandle,0,(char*)smallBuffer);
	ERRCHECK(result);

	result = codec->fileread(codec->filehandle,(char*)smallBuffer,4,bytesread,0);
	

	if(!(smallBuffer[1]=='S' &&
		 smallBuffer[2]=='I' && smallBuffer[3]=='D')) //not a sid song
	{
		delete bytesread;
		delete m;
		delete smallBuffer;
		return FMOD_ERR_FORMAT;
	}

	
	delete smallBuffer;
	signed short* myBuffer;
	myBuffer = new signed short[codec->filesize];

	//rewind file pointer
	result = codec->fileseek(codec->filehandle,0,(signed short*)myBuffer);
	ERRCHECK(result);

	//read whole file to memory
	result = codec->fileread(codec->filehandle,(signed short*)myBuffer,codec->filesize,bytesread,0);
	ERRCHECK(result);

	m->tune = new SidTune(0);
	m->tune->read((uint_least8_t*)myBuffer,codec->filesize);
	delete bytesread;
	delete myBuffer;
 
	m->m_cfg = m->config();
	// Configure default settings
	m->m_cfg.clockDefault    = SID2_CLOCK_CORRECT;
	m->m_cfg.clockForced     = false;
	m->m_cfg.clockSpeed      = SID2_CLOCK_CORRECT;
	m->m_cfg.forceDualSids   = false;
	m->m_cfg.emulateStereo   = true;
	m->m_cfg.forceDualSids   = true;
	m->m_cfg.frequency       = 44100;
	m->m_cfg.optimisation    = SID2_DEFAULT_OPTIMISATION;
	m->m_cfg.playback        = sid2_stereo;
	m->m_cfg.precision       = 16;
	m->m_cfg.sidDefault      = SID2_MODEL_CORRECT;
	m->m_cfg.sidEmulation    = NULL;
	m->m_cfg.sidModel        = SID2_MODEL_CORRECT;
	m->m_cfg.sidSamples      = true;
	m->m_cfg.leftVolume      = 255;
	m->m_cfg.rightVolume     = 255;
	
	/* Environment Modes
	sid2_envPS = Playsid
	sid2_envTP = Sidplay  - Transparent Rom
	sid2_envBS = Sidplay  - Bankswitching
	sid2_envR  = Sidplay2 - Real C64 Environment
	*/
	m->m_cfg.environment = sid2_envR;
	

	// add ReSID sidbuilder
	
	m->rs = new ReSIDBuilder("ReSID");
	m->rs->create(2);

	m->m_cfg.sidEmulation = m->rs;
	m->rs->filter(true);
	m->rs->sampling(m->m_cfg.frequency);

	m->config(m->m_cfg);


	SidTuneInfo s;
	m->tune->getInfo(s);
	m->subsongs = s.songs;
	
	m->sidwaveformat.format       = FMOD_SOUND_FORMAT_PCM16;
	m->sidwaveformat.channels     = 2;
	m->sidwaveformat.frequency    = 44100;
	m->sidwaveformat.blockalign   = (16 >> 3) * m->sidwaveformat.channels;
	m->sidwaveformat.lengthpcm    = 0xffffffff; //inifite length

	m->tune->selectSong(0);
	m->load(m->tune);

	codec->waveformat   = &(m->sidwaveformat);
	codec->numsubsounds = 0;
	codec->plugindata   = m;                    //user data value

	return FMOD_OK;
}

FMOD_RESULT F_CALLBACK sidread(FMOD_CODEC_STATE *codec, void *buffer, unsigned int size, unsigned int *read)
{
	sid *m = (sid*)codec->plugindata;
	*read = m->play((signed short*)buffer,codec->waveformat->blockalign*sizeof(signed short));
	return FMOD_OK; 
}

FMOD_RESULT F_CALLBACK sidsetposition(FMOD_CODEC_STATE *codec, int subsound, unsigned int position, FMOD_TIMEUNIT postype)
{
	sid *m = (sid*)codec->plugindata;
	if(postype==FMOD_TIMEUNIT_MS)
	{
		//can't seek at the moment
	}
	else if(postype==FMOD_TIMEUNIT_MODPATTERN)
	{
		m->tune->selectSong(position+1);
		m->load(m->tune);
	}

	return FMOD_OK;
}

FMOD_RESULT F_CALLBACK sidclose(FMOD_CODEC_STATE *codec)
{
	delete (sid*)codec->plugindata;
    return FMOD_OK;
}
FMOD_RESULT F_CALLBACK sidgetlength(FMOD_CODEC_STATE *codec, unsigned int *length, FMOD_TIMEUNIT lengthtype)
{
	sid *m = (sid*)codec->plugindata;

	if(lengthtype==FMOD_TIMEUNIT_MS)
	{
		*length = m->sidwaveformat.lengthpcm;
	}
	else if(lengthtype==FMOD_TIMEUNIT_MODPATTERN)
	{
		*length = m->subsongs;
	}
    return FMOD_OK;
}
FMOD_RESULT F_CALLBACK sidgetposition(FMOD_CODEC_STATE *  codec, unsigned int *position, FMOD_TIMEUNIT postype)
{
	sid *m = (sid*)codec->plugindata;

	if(postype==FMOD_TIMEUNIT_MS)
	{
		*position = 0;
	}
	else if(postype==FMOD_TIMEUNIT_MODPATTERN)
	{
		SidTuneInfo s;
		m->tune->getInfo(s);
		*position = s.currentSong;
	}
    return FMOD_OK;
}
